/* Copyright (C) 2008 Free Software Foundation, Inc.
   This file is part of the GNU C Library.

   The GNU C Library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2.1 of the License, or (at your option) any later version.

   The GNU C Library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with the GNU C Library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307 USA.  */

#ifndef _BITS_ATOMIC_H
#define _BITS_ATOMIC_H	1

#include <stdint.h>

/* Coldfire has no atomic compare-and-exchange operation, but the
   kernel provides userspace atomicity operations.  Use them.  */

typedef int32_t atomic32_t;
typedef uint32_t uatomic32_t;
typedef int_fast32_t atomic_fast32_t;
typedef uint_fast32_t uatomic_fast32_t;

typedef intptr_t atomicptr_t;
typedef uintptr_t uatomicptr_t;
typedef intmax_t atomic_max_t;
typedef uintmax_t uatomic_max_t;

/* The only basic operation needed is compare and exchange.  */
#define atomic_compare_and_exchange_val_acq(mem, newval, oldval)	\
  ({									\
    /* Use temporary variables to workaround call-clobberness of */	\
    /* the registers.  */						\
    __typeof (mem) _mem = mem;						\
    __typeof (oldval) _oldval = oldval;					\
    __typeof (newval) _newval = newval;					\
    register __typeof (mem) _a0 asm ("a0") = _mem;			\
    register __typeof (oldval) _d0 asm ("d0") = _oldval;		\
    register __typeof (newval) _d1 asm ("d1") = _newval;		\
    void *ptr;								\
									\
    asm ("movel #_GLOBAL_OFFSET_TABLE_@GOTPC, %2\n\t"			\
	 "lea (-6, %%pc, %2), %2\n\t"					\
	 "movel __vdso_atomic_cmpxchg_32@GOT(%2), %2\n\t"		\
	 "movel (%2), %2\n\t"						\
	 "jsr (%2)\n\t"							\
	 : "+m" (*_a0), "+d" (_d0), "=&a" (ptr)				\
	 : "d" (_d1));							\
    _d0;								\
  })

#define atomic_full_barrier()				\
  ({							\
    void * ptr;						\
							\
    asm ("movel #_GLOBAL_OFFSET_TABLE_@GOTPC, %0\n\t"	\
	 "lea (-6, %pc, %0), %0\n\t"			\
	 "movel __vdso_atomic_barrier@GOT(%0), %0\n\t"	\
	 "movel (%0), %0\n\t"				\
	 "jsr (%0)\n\t"					\
	 : "=&a" (ptr));				\
  })

#endif
